const char* QN_MLP_ThreadFl3_rcsid =
    "$Header: /u/drspeech/repos/quicknet2/QN_MLP_ThreadFl3.cc,v 1.3 2005/06/01 01:52:29 davidj Exp $";

/* Must include the config.h file first */
#include <QN_config.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include "QN_types.h"
#include "QN_Logger.h"
#include "QN_MLP_ThreadFl3.h"
#include "QN_fltvec.h"


// A simple minimum routine
static inline unsigned int
min(unsigned int a, unsigned int b)
{
    if (a<b)
	return a;
    else
	return b;
}

#ifdef QN_HAVE_LIBPTHREAD

// The actual thread worker routine, defined below
extern "C" {
	static void* worker_wrapper(void*);
};

QN_MLP_ThreadFl3::QN_MLP_ThreadFl3(int a_debug,
				   const char* a_dbgname,
				   size_t a_input, size_t a_hidden,
				   size_t a_output,
				   enum QN_OutputLayerType a_outtype,
				   size_t a_size_bunch,
				   size_t a_threads)
    : clog(a_debug, "QN_MLP_ThreadFl3", a_dbgname),
      n_input(a_input),
      n_hidden(a_hidden),
      n_output(a_output),
      n_in2hid(a_input * a_hidden),
      n_hid2out(a_hidden * a_output),
      size_input(a_input * a_size_bunch),
      size_hidden(a_hidden * a_size_bunch),
      size_output(a_output * a_size_bunch),
      out_layer_type(a_outtype),
      size_bunch(a_size_bunch),
      num_threads(a_threads)
{
    int ec;

    // Maybe we do not support all output layer types
    switch(out_layer_type)
    {
    case QN_OUTPUT_SIGMOID:
    case QN_OUTPUT_SIGMOID_XENTROPY:
    case QN_OUTPUT_SOFTMAX:
    case QN_OUTPUT_LINEAR:
	break;
    default:
	assert(0);		// Only the above output layer types are
				// supported.
    }
    // Create weights and biases
    in2hid = new float [n_in2hid + CACHE_PAD];
    hid_bias = new float [size_hidden + CACHE_PAD];
    hid2out = new float [n_hid2out + CACHE_PAD];
    out_bias = new float [size_output + CACHE_PAD];

    // Create arrays of pointers to thread-specific weights and biases
    in2hid_deltas = new float* [num_threads];
    hid_bias_deltas = new float* [num_threads];
    hid2out_deltas = new float* [num_threads];
    out_bias_deltas = new float* [num_threads];

    // Cannot have more threads than bunch size.
    assert(num_threads<=size_bunch);
    threads = new pthread_t[num_threads];

    // Set up the "action" variable/mutex/cv worker thread command channel
    action_seq = 0;
    ec = pthread_mutex_init(&action_mutex, NULL);
    if (ec)
        clog.error("failed to init action_mutex");
    ec = pthread_cond_init(&action_cv, NULL);
    if (ec)
        clog.error("failed to init action_cv");
    done_count = 0;
    ec = pthread_mutex_init(&done_mutex, NULL);
    if (ec)
        clog.error("failed to init done_mutex");
    ec = pthread_cond_init(&done_cv, NULL);
    if (ec)
        clog.error("failed to init done_cv");

    // Create threads, make them joinable
    pthread_attr_t attr;
    ec = pthread_attr_init(&attr);
    assert(ec==0);
    ec = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    assert(ec==0);
    worker_args = new QN_MLP_ThreadFl3_WorkerArg[num_threads];
    size_t i;
    for (i=0; i<num_threads; i++)
    {
	worker_args[i].threadno = i;
	worker_args[i].mlp = this;
	clog.log(QN_LOG_PER_RUN,"Creating thread %lu.", i);
	ec = pthread_create(&threads[i], &attr, worker_wrapper,
			    (void*) &worker_args[i]);
	if (ec)
	{
	    clog.error("failed to create thread number %lu - %s.",
		       i, strerror(ec));
	}
    }

    clog.log(QN_LOG_PER_RUN,"Created net, n_input=%lu n_hidden=%lu "
	     "n_output=%lu.", n_input, n_hidden, n_output);
    clog.log(QN_LOG_PER_RUN, "bunchsize=%lu size_input=%lu "
	     "size_hidden=%lu size_output=%lu.\n",
	     size_bunch, size_input, size_hidden, size_output);
    clog.log(QN_LOG_PER_RUN, "num_threads=%lu.", num_threads);
}

QN_MLP_ThreadFl3::~QN_MLP_ThreadFl3()
{
    int ec;

    clog.log(QN_LOG_PER_RUN,"Terminating threads.");
    tell_workers("destructor", ACTION_EXIT);
    wait_on_workers("destructor");

    // Wait for all of the worker threads to die
    size_t i;
    for (i = 0; i<num_threads; i++)
    {
	clog.log(QN_LOG_PER_RUN,"Waiting for end of thread %lu.", i);
        ec = pthread_join(threads[i], NULL);
	if (ec)
	    clog.error("failed to join thread %lu - %s.", i, strerror(ec));
    }
    delete[] worker_args;

    // Kill thead-related variables
    ec = pthread_cond_destroy(&done_cv);
    if (ec)
        clog.error("failed to destroy done_cv.");
    ec = pthread_mutex_destroy(&done_mutex);
    if (ec)
        clog.error("failed to destroy done_mutex.");
    ec = pthread_cond_destroy(&action_cv);
    if (ec)
        clog.error("failed to destroy action_cv.");
    ec = pthread_mutex_destroy(&action_mutex);
    if (ec)
        clog.error("failed to destroy action_mutex.");

    delete [] threads;

    delete [] out_bias_deltas;
    delete [] hid2out_deltas;
    delete [] hid_bias_deltas;
    delete [] in2hid_deltas;

    delete [] out_bias;
    delete [] hid2out;
    delete [] hid_bias;
    delete [] in2hid;
}

size_t
QN_MLP_ThreadFl3::size_layer(QN_LayerSelector layer) const
{
    size_t r;

    switch(layer)
    {
    case QN_MLP3_INPUT:
	r = n_input;
	break;
    case QN_MLP3_HIDDEN:
	r = n_hidden;
	break;
    case QN_MLP3_OUTPUT:
	r = n_output;
	break;
    default:
	r = 0;
	break;
    }
    return r;
}


void
QN_MLP_ThreadFl3::size_section(QN_SectionSelector section, size_t* output_p,
			      size_t* input_p) const
{
    switch(section)
    {
    case QN_MLP3_INPUT2HIDDEN:
	*output_p = n_hidden;
	*input_p = n_input;
	break;
    case QN_MLP3_HIDDENBIAS:
	*output_p = n_hidden;
	*input_p = 1;
	break;
    case QN_MLP3_HIDDEN2OUTPUT:
	*output_p = n_output;
	*input_p = n_hidden;
	break;
    case QN_MLP3_OUTPUTBIAS:
	*output_p = n_output;
	*input_p = 1;
	break;
    default:
	assert(0);
    }
}

size_t
QN_MLP_ThreadFl3::num_connections() const
{
    // Note how we take account of the biases separately
    return (n_input+n_output) * n_hidden + n_hidden + n_output;
}

void
QN_MLP_ThreadFl3::forward(size_t n_frames, const float* in, float* out)
{
    size_t i;
    size_t frames_this_bunch;	// Number of frames to handle this bunch

    for (i=0; i<n_frames; i += size_bunch)
    {
	frames_this_bunch = min(size_bunch, n_frames - i);
	
	forward_bunch(frames_this_bunch, in, out);
	in += n_input * frames_this_bunch;
	out += n_output * frames_this_bunch;
    }
}

void
QN_MLP_ThreadFl3::train(size_t n_frames, const float* in, const float* target,
		    float* out)
{
    size_t i;
    size_t frames_this_bunch;	// Number of frames to handle this bunch

    for (i=0; i<n_frames; i+= size_bunch)
    {
	frames_this_bunch = min(size_bunch, n_frames - i);
	train_bunch(frames_this_bunch, in, target, out);
	in += n_input * frames_this_bunch;
	out += n_output * frames_this_bunch;
	target += n_output * frames_this_bunch;
    }
}

void
QN_MLP_ThreadFl3::forward_bunch(size_t n_frames, const float* in, float* out)
{
    clog.log(QN_LOG_PER_BUNCH, "forward_bunch sending ACTION_FORWARD.");
    action_n_frames = n_frames;
    action_in = in;
    action_out = out;
    tell_workers("forward_bunch", ACTION_FORWARD);
    // ....workers work...
    wait_on_workers("forward_bunch");
    clog.log(QN_LOG_PER_BUNCH, "forward_bunch workers claim they are done.");
}

void
QN_MLP_ThreadFl3::train_bunch(size_t n_frames, const float *in,
			     const float* target, float* out)
{

    clog.log(QN_LOG_PER_BUNCH, "train_bunch sending ACTION_TRAIN.");
    action_n_frames = n_frames;
    action_in = in;
    action_out = out;
    action_target = target;
    tell_workers("train_bunch", ACTION_TRAIN);
    // ....workers work...
    wait_on_workers("train_bunch");
    clog.log(QN_LOG_PER_BUNCH, "train_bunch workers claim they are done.");

    // He we merge all of the deltas.
    // Note we want these in separate loops for better cache usage.
    size_t thread;
    for (thread = 0; thread<num_threads; thread++)
    {
	qn_mulacc_vff_vf(n_output, out_bias_deltas[thread], neg_out_learnrate,
			 out_bias);
    }
    for (thread = 0; thread<num_threads; thread++)
    {
	qn_acc_vf_vf(n_hid2out, hid2out_deltas[thread], hid2out);
    }
    for (thread = 0; thread<num_threads; thread++)
    {
	qn_mulacc_vff_vf(n_hidden, hid_bias_deltas[thread], neg_hid_learnrate,
			 hid_bias);
    }
    for (thread = 0; thread<num_threads; thread++)
    {
	qn_acc_vf_vf(n_in2hid, in2hid_deltas[thread], in2hid);
    }
}


void
QN_MLP_ThreadFl3::set_weights(enum QN_SectionSelector which,
			     size_t row, size_t col,
			     size_t n_rows, size_t n_cols,
			     const float* from)
{
    float* start;		// The first element of the relveant weight
				// matrix that we want
    size_t total_cols;		// The total number of columns in the given
				// weight matrix

    start = findweights(which, row, col, n_rows, n_cols,
			&total_cols);
    clog.log(QN_LOG_PER_SUBEPOCH,
	     "Set weights %s @ (%lu,%lu) size (%lu,%lu).",
	     nameweights(which), row, col, n_rows, n_cols);
    qn_copy_mf_smf(n_rows, n_cols, total_cols, from, start);
}

void
QN_MLP_ThreadFl3::get_weights(enum QN_SectionSelector which,
			      size_t row, size_t col,
			      size_t n_rows, size_t n_cols,
			      float* to) const
{
    float* start;		// The first element of the relveant weight
				// matrix that we want
    size_t total_cols;		// The total number of columns in the given
				// weight matrix
    start = findweights(which, row, col, n_rows, n_cols, &total_cols);
    clog.log(QN_LOG_PER_SUBEPOCH,
	     "Get weights %s @ (%lu,%lu) size (%lu,%lu).",
	     nameweights(which), row, col, n_rows, n_cols);
    qn_copy_smf_mf(n_rows, n_cols, total_cols, start, to);
}

void
QN_MLP_ThreadFl3::set_learnrate(enum QN_SectionSelector which, float learnrate)
{
    switch(which)
    {
    case QN_MLP3_INPUT2HIDDEN:
	neg_in2hid_learnrate = -learnrate;
	break;
    case QN_MLP3_HIDDEN2OUTPUT:
	neg_hid2out_learnrate = -learnrate;
	break;
    case QN_MLP3_HIDDENBIAS:
	neg_hid_learnrate = -learnrate;
	break;
    case QN_MLP3_OUTPUTBIAS:
	neg_out_learnrate = -learnrate;
	break;
    default:
	assert(0);
	break;
    }
}

float 
QN_MLP_ThreadFl3::get_learnrate(enum QN_SectionSelector which) const
{
    float res;			// Returned learning rate

    switch(which)
    {
    case QN_MLP3_INPUT2HIDDEN:
	res = -neg_in2hid_learnrate;
	break;
    case QN_MLP3_HIDDEN2OUTPUT:
	res = -neg_hid2out_learnrate;
	break;
    case QN_MLP3_HIDDENBIAS:
	res = -neg_hid_learnrate;
	break;
    case QN_MLP3_OUTPUTBIAS:
	res = -neg_out_learnrate;
	break;
    default:
	res = 0;
	assert(0);
	break;
    }
    return res;
}

const char*
QN_MLP_ThreadFl3::nameweights(QN_SectionSelector which)
{
    switch(which)
    {
    case QN_MLP3_INPUT2HIDDEN:
	return "input2hidden";
	break;
    case QN_MLP3_HIDDEN2OUTPUT:
	return "hidden2output";
	break;
    case QN_MLP3_HIDDENBIAS:
	return "hiddenbias";
	break;
    case QN_MLP3_OUTPUTBIAS:
	return "outputbias";
	break;
    default:
	assert(0);
	return "unknown";
	break;
    }
}

float*
QN_MLP_ThreadFl3::findweights(QN_SectionSelector which,
			     size_t row, size_t col,
			     size_t n_rows, size_t n_cols,
			     size_t* total_cols_p) const
{
// Note - for a weight matrix, inputs are across a row, outputs are
// ..down a column

    float *weights;		// Pointer to bit of weight matrix requested
    size_t total_rows;		// The number of rows in the selected matrix
    size_t total_cols;		// The number of cols in the selected matrix

    size_section(which, &total_rows, &total_cols);
    switch(which)
    {
    case QN_MLP3_INPUT2HIDDEN:
	weights = in2hid;
	break;
    case QN_MLP3_HIDDEN2OUTPUT:
	weights = hid2out;
	break;
    case QN_MLP3_HIDDENBIAS:
	weights = hid_bias;
	break;
    case QN_MLP3_OUTPUTBIAS:
	weights = out_bias;
	break;
    default:
	weights = NULL;
	assert(0);
    }
    assert(row<total_rows);
    assert(row+n_rows<=total_rows);
    assert(col<total_cols);
    assert(col+n_cols<=total_cols);
    weights += col + row * total_cols;
    *total_cols_p = total_cols;
    return(weights);
}

size_t
QN_MLP_ThreadFl3::get_bunchsize()
{
    return size_bunch;
}

void
QN_MLP_ThreadFl3::tell_workers(const char* logstr, enum Action action)
{
    int ec;

    // Send an action to all of the worker threads
    ec = pthread_mutex_lock(&action_mutex);
    if (ec)
    {
        clog.error("%s - failed to lock action_mutex in tell_workers - %s",
		   logstr, strerror(ec));
    }
    action_command = action;
    action_seq++;		// Increment sequence to indicate new command
    ec = pthread_cond_broadcast(&action_cv);
    if (ec)
    {
        clog.error("%s - failed to unlock action_mutex in tell_workers - %s",
		   logstr, strerror(ec));
    }
    ec = pthread_mutex_unlock(&action_mutex);
    if (ec)
    {
        clog.error("%s - failed to unlock action_mutex in tell_workers - %s",
		   logstr, strerror(ec));
    }
}

void
QN_MLP_ThreadFl3::wait_on_workers(const char* logstr)
{
    int ec;

    // Send an action to all of the worker threads
    ec = pthread_mutex_lock(&done_mutex);
    if (ec)
    {
        clog.error("%s - failed to lock done_mutex in wait_on_workers - %s",
		   logstr, strerror(ec));
    }
    while (done_count < num_threads)
    {
	ec = pthread_cond_wait(&done_cv, &done_mutex);
	if (ec)
	{
	    clog.error("%s - failed wait on done_cv in wait_on_workers - %s",
		       logstr, strerror(ec));
	}

    }
    // Reset count for next time around whilst we have the mutex
    done_count = 0;
    ec = pthread_mutex_unlock(&done_mutex);
    if (ec)
    {
        clog.error("%s - failed to unlock done_mutex in wait_on_workers - %s",
		   logstr, strerror(ec));
    }
}

void
QN_MLP_ThreadFl3::worker(size_t threadno)
{
    int ec;			// Error code
    enum Action action;		// Our copy of the action we are to do
    unsigned int last_action_seq = 0; // Last action sequence no.
    size_t i;			// Counter
    int exiting = 0;		// Set to true when exiting.

    clog.log(QN_LOG_PER_RUN, "Thread %d up and self-aware", threadno);


    // Sized of our part of bunch.
    const size_t w_num_threads = num_threads;
    const size_t w_size_bunch = \
	(size_bunch + w_num_threads - 1)/ w_num_threads;
    const size_t w_first_frame = threadno * w_size_bunch;
    // Our copies of important net params
    const size_t w_n_input = n_input;
    const size_t w_n_hidden = n_hidden;
    const size_t w_n_output = n_output;
    const size_t w_n_in2hid = n_in2hid;
    const size_t w_n_hid2out = n_hid2out;
    // The size of our local matrices
    const size_t w_size_input = w_n_input * w_size_bunch;
    const size_t w_size_hidden = w_n_hidden * w_size_bunch;
    const size_t w_size_output = w_n_output * w_size_bunch;
    // Offsets for reading inputs and writing results
    const size_t w_input_offset = w_size_input * threadno;
    const size_t w_output_offset = w_size_output * threadno;
    // Other state
    const enum QN_OutputLayerType w_out_layer_type = out_layer_type;

    // Allocate our local state
    float* w_in2hid_delta = new float [w_n_in2hid + CACHE_PAD];
    in2hid_deltas[threadno] = w_in2hid_delta;
    float* w_hid_x = new float [w_size_hidden + CACHE_PAD];
    float* w_hid_y = new float [w_size_hidden + CACHE_PAD];
    float* w_hid2out_delta = new float [w_n_hid2out + CACHE_PAD];
    hid2out_deltas[threadno] = w_hid2out_delta;
    float* w_out_x = new float [w_size_output + CACHE_PAD];

    float* w_out_dedx = new float [w_size_output + CACHE_PAD];
    float* w_out_bias_delta = new float [w_n_output + CACHE_PAD];
    out_bias_deltas[threadno] = w_out_bias_delta;

    // These intermediate values are only needed for sigmoid output layers
    // ...and only then when we are NOT doing cross entropy
    float* w_out_dedy;
    float* w_out_dydx;
    if (w_out_layer_type==QN_OUTPUT_SIGMOID)
    {
	w_out_dedy = new float [w_size_output + CACHE_PAD];
	w_out_dydx = new float [w_size_output + CACHE_PAD];
    }
    float* w_hid_dedy = new float [w_size_hidden + CACHE_PAD];
    float* w_hid_dydx = new float [w_size_hidden + CACHE_PAD];
    float* w_hid_dedx = new float [w_size_hidden + CACHE_PAD];
    float* w_hid_bias_delta = new float [w_n_hidden + CACHE_PAD];
    hid_bias_deltas[threadno] = w_hid_bias_delta;

    size_t w_cur_frames;	// Number of frames this time around.
    const float* w_in;		// Input frame pointer
    float* w_out;		// Output frame pointer
    const float* w_target;	// Target frame pointer
    size_t w_cur_size_hidden;	// Hidden size this time around
    size_t w_cur_size_output;	// Output size this time around
    float* w_out_x_p;		// Pointer into w_out_x
    float* w_out_p;		// Pointer into w_out

    // The main worker loop
    // Exited by a call to pthread_exit()
    while(!exiting)
    {
	//  Wait to be told what to do by sitting on the action
	// ...condition variable waiting for action_seq to change
	ec = pthread_mutex_lock(&action_mutex);
	if (ec)
	    clog.error("failed to lock action_mutex in worker");
	clog.log(QN_LOG_PER_BUNCH, "Thread %d waiting.", threadno); //??
	while (action_seq == last_action_seq)
	{
	    ec = pthread_cond_wait(&action_cv, &action_mutex);
	    if (ec)
		clog.error("failed to wait on action_cv in worker");
	}
	action = action_command;
	last_action_seq = action_seq;
	ec = pthread_mutex_unlock(&action_mutex);
	if (ec)
	    clog.error("failed to unlock action_mutex in worker");

	switch(action)
	{
	case ACTION_EXIT:
	    exiting = 1;
	    break;
	case ACTION_FORWARD:
	    // Work out how many frames we deal with.
	    if (action_n_frames>w_first_frame)
	    {
		w_cur_frames = \
		    min(action_n_frames - w_first_frame, w_size_bunch);
	    }
	    else
		w_cur_frames = 0;
	    clog.log(QN_LOG_PER_BUNCH, "Thread %d forward bunch %lu frames.",
		     threadno, w_cur_frames);
	    if (w_cur_frames>0)
	    {
		// Note that for forward we use non-thread weights in-place

		w_cur_size_hidden = w_n_hidden * w_cur_frames;
		w_cur_size_output = w_n_output * w_cur_frames;
		w_in = action_in + w_input_offset;
		w_out = action_out + w_output_offset;

		// First layer
		qn_copy_vf_mf(w_cur_frames, w_n_hidden, hid_bias, w_hid_x);
		qn_mulntacc_mfmf_mf(w_cur_frames, w_n_input, w_n_hidden,
				    w_in, in2hid, w_hid_x);
		qn_sigmoid_vf_vf(w_cur_size_hidden, w_hid_x, w_hid_y);

		// Second layer
		qn_copy_vf_mf(w_cur_frames, w_n_output, out_bias, w_out_x);
		qn_mulntacc_mfmf_mf(w_cur_frames, w_n_hidden, w_n_output,
				    w_hid_y, hid2out, w_out_x);
		switch(w_out_layer_type)
		{
		case QN_OUTPUT_SIGMOID:
		case QN_OUTPUT_SIGMOID_XENTROPY:
		    qn_sigmoid_vf_vf(w_cur_size_output, w_out_x, w_out);
		    break;
		case QN_OUTPUT_SOFTMAX:
		    w_out_x_p = w_out_x;
		    w_out_p = w_out;
		    
		    for (i=0; i<w_cur_frames; i++)
		    {
			qn_softmax_vf_vf(w_n_output, w_out_x_p, w_out_p);
			w_out_x_p += w_n_output;
			w_out_p += w_n_output;
		    }
		    break;
		case QN_OUTPUT_LINEAR:
		    qn_copy_vf_vf(w_cur_size_output, w_out_x, w_out);
		    break;
		default:
		    assert(0);
		}
	    }
	    break;
	case ACTION_TRAIN:
	    // Work out how many frames we deal with.
	    if (action_n_frames>w_first_frame)
	    {
		w_cur_frames = \
		    min(action_n_frames - w_first_frame, w_size_bunch);
	    }
	    else
		w_cur_frames = 0;
	    clog.log(QN_LOG_PER_BUNCH, "Thread %d train bunch %lu frames.",
		     threadno, w_cur_frames);
	    if (w_cur_frames>0)
	    {
		w_cur_size_hidden = w_n_hidden * w_cur_frames;
		w_cur_size_output = w_n_output * w_cur_frames;

		w_in = action_in + w_input_offset;
		w_out = action_out + w_output_offset;
		w_target = action_target + w_output_offset;

		// For training, need our own copy of weights and biases

		// First layer
		qn_copy_vf_mf(w_cur_frames, w_n_hidden, hid_bias, w_hid_x);

		qn_mulntacc_mfmf_mf(w_cur_frames, w_n_input, w_n_hidden,
				    w_in, in2hid, w_hid_x);
		qn_sigmoid_vf_vf(w_cur_size_hidden, w_hid_x, w_hid_y);

		// Second layer
		qn_copy_vf_mf(w_cur_frames, w_n_output, out_bias, w_out_x);

		qn_mulntacc_mfmf_mf(w_cur_frames, w_n_hidden, w_n_output,
				    w_hid_y, hid2out, w_out_x);
		switch(w_out_layer_type)
		{
		case QN_OUTPUT_SIGMOID:
		    // Forward part
		    qn_sigmoid_vf_vf(w_cur_size_output, w_out_x, w_out);
		    // Backward part
		    // For a sigmoid layer, de/dx = de/dy . dy/dx
		    qn_sub_vfvf_vf(w_cur_size_output, w_out, w_target,
				   w_out_dedy);
		    qn_dsigmoid_vf_vf(w_cur_size_output, w_out, w_out_dydx);
		    qn_mul_vfvf_vf(w_cur_size_output, w_out_dydx, w_out_dedy,
				   w_out_dedx);
		    break;
		case QN_OUTPUT_SIGMOID_XENTROPY:
		    // Forward part
		    qn_sigmoid_vf_vf(w_cur_size_output, w_out_x, w_out);
		    // Backward part
		    // For cross entropy sigmoid, dx = dy
		    qn_sub_vfvf_vf(w_cur_size_output, w_out, w_target,
				   w_out_dedx); 
		    break;
		case QN_OUTPUT_SOFTMAX:
		    // Forward part - need to do each frame individually
		    w_out_x_p = w_out_x;
		    w_out_p = w_out;
		    for (i=0; i<w_cur_frames; i++)
		    {
			qn_softmax_vf_vf(w_n_output, w_out_x_p, w_out_p);
			w_out_x_p += w_n_output;
			w_out_p += w_n_output;
		    }
		    // Backward part
		    // For cross entropy softmax, dx = dy
		    qn_sub_vfvf_vf(w_cur_size_output, w_out, w_target,
				   w_out_dedx);
		    break;
		case QN_OUTPUT_LINEAR:
		    // Forward part
		    qn_copy_vf_vf(w_cur_size_output, w_out_x, w_out);
		    // Backward part
		    // For linear, dx = dy
		    qn_sub_vfvf_vf(w_cur_size_output, w_out, w_target,
				   w_out_dedx);
		    break;
		default:
		    // Unknown layer type
		    assert(0);
		} // End of output unit switch

		// Back propagate error through hidden to output weights.
		qn_mul_mfmf_mf(w_cur_frames, w_n_output, w_n_hidden,
			       w_out_dedx, hid2out, w_hid_dedy); 

		// Update hid2out weight deltas.
		// Note learning rate is applied later.
		qn_copy_f_vf(w_n_hid2out, 0.0f, w_hid2out_delta);
		qn_multnacc_fmfmf_mf(w_cur_frames, w_n_output, w_n_hidden,
				     neg_in2hid_learnrate, w_out_dedx,
				     w_hid_y, w_hid2out_delta);
		
		// Calculate output biases delta.
		qn_sumcol_mf_vf(w_cur_frames, w_n_output, w_out_dedx,
				w_out_bias_delta);

		// Propogate error back through sigmoid
		qn_dsigmoid_vf_vf(w_cur_size_hidden, w_hid_y, w_hid_dydx);
		qn_mul_vfvf_vf(w_cur_size_hidden, w_hid_dydx, w_hid_dedy,
			       w_hid_dedx);

		// Calculate in2hid weight deltas.
		qn_copy_f_vf(w_n_in2hid, 0.0f, w_in2hid_delta);
		qn_multnacc_fmfmf_mf(w_cur_frames, w_n_hidden, w_n_input,
				     neg_in2hid_learnrate , w_hid_dedx, 
				     w_in, w_in2hid_delta);

		// Calculate bias deltas.
		qn_sumcol_mf_vf(w_cur_frames, w_n_hidden, w_hid_dedx,
				w_hid_bias_delta);

	    } // End of "if frames to do"
	    break;
	default:
	    // Unknown action
	    assert(0);
	}
	
	// Signal that we're done
	ec = pthread_mutex_lock(&done_mutex);
	if (ec)
	    clog.error("failed to lock done_mutex in worker");
	done_count++;			// Indicate one more thread is done.
	if (done_count==num_threads)
	{
	    ec = pthread_cond_signal(&done_cv);
	    if (ec)
		clog.error("failed to signal cond_cv in worker");
	}
	ec = pthread_mutex_unlock(&done_mutex);

    }
    delete [] w_hid_bias_delta;
    delete [] w_hid_dedx;
    delete [] w_hid_dydx;
    delete [] w_hid_dedy;

    if (w_out_layer_type==QN_OUTPUT_SIGMOID)
    {
	delete [] w_out_dydx;
	delete [] w_out_dedy;
    }
    delete [] w_out_bias_delta;
    delete [] w_out_dedx;

    delete [] w_out_x;
    delete [] w_hid2out_delta;
    delete [] w_hid_y;
    delete [] w_hid_x;
    delete [] w_in2hid_delta;

    clog.log(QN_LOG_PER_RUN, "Thread %d exited.", threadno);
    pthread_exit(NULL);
}

extern "C" {

static void*
worker_wrapper(void *args)
{
    QN_MLP_ThreadFl3* mlp;
    size_t threadno;

    QN_MLP_ThreadFl3_WorkerArg* unvoided_args = \
	(QN_MLP_ThreadFl3_WorkerArg*) args;
    mlp = unvoided_args->mlp;
    threadno = unvoided_args->threadno;

    mlp->worker(threadno);

    return NULL;
}

}; // extern "C"

#endif // #ifde QN_HAVE_LIBPTHREAD

